#include "GMidiin.h"



ISR(USART_RX_vect)
{
	Midiin.sendByte(UDR0);
}



GMidiin::GMidiin()
{
	GMidiin::init();
}



void GMidiin::init ()
{
	messageBuffer = (volatile unsigned char *)malloc(MESSAGE_BUFFER_SIZE);
}


void GMidiin::begin()
{
	// Set baud rate to 31,250. Requires modification if clock speed is not 16MHz.
	UBRR0H = ((F_CPU / 16 + MIDI_BAUD_RATE / 2) / MIDI_BAUD_RATE - 1) >> 8;
	UBRR0L = ((F_CPU / 16 + MIDI_BAUD_RATE / 2) / MIDI_BAUD_RATE - 1);
	
	// enable rx
	UCSR0B |= _BV(RXEN0);
	
	// USART RX interrupt enable bit on
	UCSR0B |= _BV(RXCIE0);

	messagePointer = 0;
	messageOverflow = SYSEX_NO_OVERFLOW;
}


void GMidiin::stop()
{
	UCSR0B &= ~(_BV(RXCIE0));
}


void GMidiin::setNoteOnFunc(void (*funcPtr)(int channel, int noteNo, int velocity))
{
	noteOnCallback = funcPtr;
}


void GMidiin::setNoteOffFunc(void (*funcPtr)(int channel, int noteNo, int velocity))
{
	noteOffCallback = funcPtr;
}


void GMidiin::setCcFunc(void (*funcPtr)(int channel, int ccNo, int value))
{
	ccCallback = funcPtr;
}


void GMidiin::setPrgChangeFunc(void (*funcPtr)(int channel, int prgNo))
{
	prgChangeCallback = funcPtr;
}

void GMidiin::setPitchBendFunc(void (*funcPtr)(int channel, int value))
{
	pitchBendCallback = funcPtr;
}


void GMidiin::setSysexFunc(void (*funcPtr)(volatile unsigned char *sysex, int len, int overflow))
{
	sysexCallback = funcPtr;
}



void GMidiin::sendByte (unsigned char uart_buffer)
{
	// If received byte is "status byte"
	if (uart_buffer & 0x80) {

		//If Sysex end
		if (uart_buffer == SYSEX_END_BYTE) {
			if (!messagePointer || (messageBuffer[0] != STATUS_SYSEX)) {
				messagePointer = 0;
				return;
			}
			messageBuffer[messagePointer++] = SYSEX_END_BYTE;
			sysexCallback(messageBuffer, messagePointer, messageOverflow);
			messageOverflow = SYSEX_NO_OVERFLOW;
			messagePointer = 0;
			return;
		}
		
		//else, "status byte" will be written in top of buffer
		messagePointer = 0;

	} else if (!messagePointer) {
		// When receiving a "non-status byte" at top of buffer, reuse previous
		messagePointer++;
	}
	
	messageBuffer[messagePointer++] = uart_buffer;
	

	// Set pointer back to 1 if pointer (1byte) over flows.
	// This limits the max length of Sysex to 255 bytes but is OK for now.
	if (messagePointer == MESSAGE_BUFFER_SIZE) {
		messagePointer = 1;
		messageOverflow = SYSEX_OVERFLOW;
	}
	
		

	switch (messageBuffer[0] & 0xF0) {

		case STATUS_NOTEON:
			if (messagePointer == 3) {
				noteOnCallback((messageBuffer[0] & 0x0F), messageBuffer[1], messageBuffer[2]);
				messagePointer = 0;
			}
			break;

		case STATUS_NOTEOFF:
			if (messagePointer == 3) {
				noteOffCallback((messageBuffer[0] & 0x0F), messageBuffer[1], messageBuffer[2]);
				messagePointer = 0;
			}
			break;

		case STATUS_CC:
			if (messagePointer == 3) {
				ccCallback((messageBuffer[0] & 0x0F), messageBuffer[1], messageBuffer[2]);
				messagePointer = 0;
			}
			break;

		case STATUS_PRGCHANGE:
			if (messagePointer == 2) {
				prgChangeCallback((messageBuffer[0] & 0x0F), messageBuffer[1]);
				messagePointer = 0;
			}
			break;

		case STATUS_PITCHBEND:
			if (messagePointer == 3) {
				pitchBendCallback((messageBuffer[0] & 0x0F), (messageBuffer[2] << 7) + messageBuffer[1]);
				messagePointer = 0;
			}
			break;

		default:
			return;
	}
}

GMidiin Midiin = GMidiin();


